/**
 * This file is part of Seneschal, a ranking function generator for
 * machine arithmetic.
 * <http://www.philipp.ruemmer.org/seneschal.shtml>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * Seneschal is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Seneschal is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Seneschal.  If not, see <http://www.gnu.org/licenses/>.
 */

package bitRanks

import ap.PresburgerTools
import ap.basetypes.IdealInt
import ap.terfor.{TermOrder, ConstantTerm, OneTerm, VariableTerm, Term}
import ap.terfor.linearcombination.LinearCombination
import ap.terfor.conjunctions.{Conjunction, ReduceWithConjunction, Quantifier,
                               NegatedConjunctions}
import ap.terfor.equations.{EquationConj, NegEquationConj}
import ap.terfor.inequalities.InEqConj
import ap.terfor.substitutions.{VariableSubst, ConstantSubst}
import ap.proof.ModelSearchProver
import ap.util.{Debug => ApDebug, FilterIt, Seqs}

import bitRanks.util.Debug

import scala.collection.mutable.PriorityQueue

object RankingFinder {
  
  private def AC = Debug.AC_RANKING_FINDER

  import Conjunction._

  abstract class NewConstantDef
  case class NewDivConstant(expr : LinearCombination,
                            denom : IdealInt) extends NewConstantDef
  
  //////////////////////////////////////////////////////////////////////////////
  
  /**
   * Split negated equations into disjunctions of inequalities
   */
  private def splitEqs(c : Conjunction) : Iterator[Conjunction] = {
    ////////////////////////////////////////////////////////////////////////////
    ApDebug.assertPre(RankingFinder.AC, PresburgerTools.isQFPresburgerConjunction(c))
    ////////////////////////////////////////////////////////////////////////////
    
    if (c.isFalse)
      Iterator.empty
    else    
      c.arithConj.negativeEqs.firstOption match {
        case None => Iterator.single(c)
        case Some(eq) => {
          val reducer = ReduceWithConjunction(Conjunction.TRUE, c.order)
          val conj0 = reducer(conj(Array(c, InEqConj(eq, c.order)), c.order))
          val conj1 = reducer(conj(Array(c, InEqConj(-eq, c.order)), c.order))
          splitEqs(conj0) ++ splitEqs(conj1)
        }
      }
  }

  private def splitDivisibilities(c : Conjunction) : Iterator[Conjunction] = {
    ////////////////////////////////////////////////////////////////////////////
    ApDebug.assertPre(RankingFinder.AC, PresburgerTools.isQFPresburgerConjunction(c))
    ////////////////////////////////////////////////////////////////////////////

    var order = c.order
    
    // we replace all the quantified variables with fresh constants
    val groundEquations =
      (for ((d, i) <- c.negatedConjs.elements.zipWithIndex) yield {
         ///////////////////////////////////////////////////////////////////////
         ApDebug.assertInt(RankingFinder.AC, d.isNonDivisibility)
         ///////////////////////////////////////////////////////////////////////
      
         val const = new ConstantTerm ("div_" + i)
         order = order.extend(const, Set())
      
         val subst = new VariableSubst(0, List(const), order)
         val newEq = subst(d.arithConj)

         ///////////////////////////////////////////////////////////////////////
         // there should only be one quantifier, which is no longer necessary now
         ApDebug.assertInt(RankingFinder.AC, newEq.variables.isEmpty)
         ///////////////////////////////////////////////////////////////////////

         newEq.negate
       }).toList
    
    val groundDisjunct = conj(groundEquations ++ List(c.arithConj), order)
    
    // then we quantify all the remaining constants
    // (because we are not interested in their values)
    val quantifiedDisjunct =
      quantify(Quantifier.EX, c.constants, groundDisjunct, order)
      
    // And then we compute all the satisfying assignments for this formula,
    // which describe the "interesting" instantiations of the divisibility
    // constraints. This will terminate, unless the considered program contains
    // variables with an unbounded range (which should not happen)
    for (assignment <- PresburgerTools.enumModels(quantifiedDisjunct)) yield {
      val reducer = ReduceWithConjunction(assignment, order)
      val res = reducer(groundDisjunct)
      //////////////////////////////////////////////////////////////////////////
      ApDebug.assertInt(RankingFinder.AC, res.constants subsetOf c.constants)
      //////////////////////////////////////////////////////////////////////////
      res
    }
  }

  //////////////////////////////////////////////////////////////////////////////
  
  /**
   * Set up the system of equations and inequalities that describe common linear
   * ranking functions (and their lower bounds) of the given disjuncts
   */
  private def setupRankingConstraints(disjuncts : Seq[Conjunction],
                                      prePostConsts : List[(ConstantTerm, ConstantTerm)],
                                      order : TermOrder)
                        : (List[ConstantTerm], ConstantTerm, Conjunction) = {
    val rankingCoeffs =
      for (i <- List.range(0, prePostConsts.size)) yield new ConstantTerm ("coeff_" + i)
    val lowerBound = new ConstantTerm ("lowerBound")
    
    val newOrder =
      (TermOrder.EMPTY /: (rankingCoeffs ++ List(lowerBound)))((o, l) => o.extend(l, Set()))
    val reducer = ReduceWithConjunction(Conjunction.TRUE, newOrder)
    
    val constraint =
        conj(for (disjunct <- disjuncts.elements) yield
               setupRankingConstraints(toInEqs(disjunct),
                                       rankingCoeffs, lowerBound,
                                       prePostConsts,
                                       newOrder),
             newOrder)
    (rankingCoeffs, lowerBound, reducer(constraint))
  }

  /**
   * Set up the system of equations and inequalities that describe linear
   * ranking functions (and their lower bounds) of the given system of
   * inequalities over <code>prePostConsts</code>.
   */
  private def setupRankingConstraints(inEqs : Seq[LinearCombination],
                                      rankingCoeffs : List[ConstantTerm],
                                      lowerBound : ConstantTerm,
                                      prePostConsts : List[(ConstantTerm, ConstantTerm)],
                                      order : TermOrder) : Conjunction = {
    ////////////////////////////////////////////////////////////////////////////
    ApDebug.assertPre(RankingFinder.AC, rankingCoeffs.size == prePostConsts.size)
    ////////////////////////////////////////////////////////////////////////////
    
    val lambda1 = for (i <- List.range(0, inEqs.size)) yield new ConstantTerm ("lambda_1_" + i)
    val lambda2 = for (i <- List.range(0, inEqs.size)) yield new ConstantTerm ("lambda_2_" + i)

    val newOrder = (order /: (lambda1 ++ lambda2))((o, l) => o.extend(l, Set()))
    
    // lambda1 * A' = 0
    val eqs1 = for ((_, postConst) <- prePostConsts) yield
      LinearCombination(for ((l1, lc) <- lambda1.elements zip inEqs.elements)
                        yield (lc get postConst, l1),
                        newOrder)
    
    // (lambda1 - lambda2) * A = 0
    val eqs2 = for ((preConst, _) <- prePostConsts) yield
      LinearCombination(for (((l1, l2), lc) <- lambda1.elements zip
                                               lambda2.elements zip
                                               inEqs.elements;
                             val coeff = lc get preConst;
                             l <- Iterator.fromValues((coeff, l1), (-coeff, l2)))
                        yield l,
                        newOrder)

    // lambda2 * (A + A') = 0
    val eqs3 = for ((preConst, postConst) <- prePostConsts) yield
      LinearCombination(for ((l2, lc) <- lambda2.elements zip inEqs.elements)
                        yield ((lc get preConst) + (lc get postConst), l2),
                        newOrder)

    // lambda2 * b < 0
    val inEqLC2 =
      LinearCombination(for ((l2, lc) <- lambda2.elements zip inEqs.elements)
                        yield (lc.constant, l2),
                        newOrder)
    val inEqLhs2 = inEqLC2.+(-LinearCombination.ONE)(newOrder)
    
    // the coefficients of the ranking function 
    // lambda2 * A'
    val coeffEqs =
      for (((_, postConst), coeff) <- prePostConsts zip rankingCoeffs) yield
        LinearCombination((for ((l2, lc) <- lambda2.elements zip inEqs.elements)
                           yield (lc get postConst, l2)) ++
                          (Iterator single (IdealInt.MINUS_ONE, coeff)),
                          newOrder)

    // the lower bound
    // -lambda1 * b
    val inEqLC1 =
      LinearCombination(for ((l1, lc) <- lambda1.elements zip inEqs.elements)
                        yield (lc.constant, l1),
                        newOrder)
    val inEqLhs1 = inEqLC1.+(-LinearCombination(lowerBound, newOrder))(newOrder)
    
    // all the lambdas are non-negative
    val nonNegLambdas =
      for (l <- lambda1 ++ lambda2) yield LinearCombination(l, newOrder)
    
    val eqsConj = EquationConj(eqs1 ++ eqs2 ++ eqs3 ++ coeffEqs, newOrder)
    val inEqConj = InEqConj(List(inEqLhs2, inEqLhs1) ++ nonNegLambdas, newOrder)
    
    val unquantifiedRes = Conjunction.conj(Array(eqsConj, inEqConj), newOrder)
    
    quantify(Quantifier.EX, lambda1 ++ lambda2, unquantifiedRes, newOrder)
  }

  /**
   * Convert an arithmetic conjunction into a conjunction
   * <code>t1 <= 0 & t2 <= 0 & ...</code> of inequalities
   */
  private def toInEqs(c : Conjunction) : Seq[LinearCombination] = {
    ////////////////////////////////////////////////////////////////////////////
    ApDebug.assertPre(RankingFinder.AC,
                      PresburgerTools.isQFPresburgerConjunction(c) &&
                      c.negatedConjs.isTrue && c.arithConj.negativeEqs.isTrue)
    ////////////////////////////////////////////////////////////////////////////
    val ac = c.arithConj

    (for (lc <- ac.positiveEqs; lc2 <- List(lc, -lc)) yield lc2) ++
    (for (lc <- ac.inEqs) yield -lc)
  }

  //////////////////////////////////////////////////////////////////////////////
  
  /**
   * Determine a (more or less) minimal solution of the given constraint,
   * in terms of the absolute values of the solution components
   */
  private def minimalSolution(c : Conjunction) : Conjunction = {
    val order = c.order
    print("Solving ... ")
    val model = ModelSearchProver(c.negate, order)
    
    if (model.isFalse) {
      // nothing to optimise
      println("no solution")
      Conjunction.FALSE
    } else {
      println("found a solution")      
      print("Minimising the solution ... ")
      
      var bestSolution = model
      
      val consts = c.constants.toList
      val prelimValues = for (const <- consts) yield extractValue(model, const)
      
      // for the time being, we just assume that the solution determines the
      // values of all constants (which should be reasonable)
      val values = for (v <- prelimValues) yield v.get

      // we construct bounds on the values of the constants (based on the
      // initial solution) and successively decrease the bounds to get smaller
      // solutions
      val currentBounds = (for (v <- values) yield v.abs).toArray
      
      implicit def orderPair(x : (IdealInt, Int)) : Ordered[(IdealInt, Int)] =
        new Ordered[(IdealInt, Int)] {
          def compare(y : (IdealInt, Int)) : Int = (x _1) compareAbs (y _1)
        }
      
      // the bounds are organised in a priority queue: we always try to
      // decrease the greatest bound up to this point
      val boundQueue = new PriorityQueue[(IdealInt, Int)]
      boundQueue ++= currentBounds.elements.zipWithIndex
      
      while (!boundQueue.isEmpty && (boundQueue.max _1) >= 2) {
        val (oldBound, index) = boundQueue.dequeue
        currentBounds(index) = oldBound / 2
        
        val boundConstraints =
          conj(for ((const, bound) <- consts.elements zip currentBounds.elements)
               yield boundValues(const, bound, order),
               order)
        
        // check whether the current solution already satisfies the new bounds
        val newModel =
          if (ReduceWithConjunction(bestSolution, order)(boundConstraints).isTrue) {
            bestSolution
          } else {
            val completeConstraint = conj(Array(c, boundConstraints), order)
            ModelSearchProver(completeConstraint.negate, order)
          }
        
        if (newModel.isFalse) {
          // go back to the old bound and do not try to change this bound again
          currentBounds(index) = oldBound
        } else {
          // we will keep this new bound and continue
          boundQueue += ((oldBound / 2, index))
          bestSolution = newModel
        }
      }
      
      println("done")
      bestSolution
    }
  }
  
  private def boundValues(const : ConstantTerm,
                          bound : IdealInt,
                          order : TermOrder) : Conjunction = {
    ////////////////////////////////////////////////////////////////////////////
    ApDebug.assertPre(RankingFinder.AC, bound.signum >= 0)
    ////////////////////////////////////////////////////////////////////////////
    val lowerBound =
      LinearCombination(Array((IdealInt.ONE, const), (bound, OneTerm)), order)
    val upperBound =
      LinearCombination(Array((IdealInt.MINUS_ONE, const), (bound, OneTerm)), order)
    conj(InEqConj(Array(lowerBound, upperBound), order), order)
  }
  
  //////////////////////////////////////////////////////////////////////////////
  
  /**
   * Extract the value of a constant from a solution (a conjunction of
   * equations). If the constant does not occur in the solution,
   * <code>None</code> is returned
   */
  private def extractValue(solution : Conjunction,
                           const : ConstantTerm) : Option[IdealInt] = {
    ////////////////////////////////////////////////////////////////////////////
    ApDebug.assertPre(RankingFinder.AC,
                      solution.predConj.isTrue &&
                      solution.negatedConjs.isTrue &&
                      solution.arithConj.inEqs.isTrue &&
                      solution.arithConj.negativeEqs.isTrue)
    ////////////////////////////////////////////////////////////////////////////

    for (entry <- solution.arithConj.positiveEqs.toMap get const) yield {
      //////////////////////////////////////////////////////////////////////////
      ApDebug.assertInt(RankingFinder.AC,
                        entry.constants == Set(const) && entry.leadingCoeff.isOne)
      //////////////////////////////////////////////////////////////////////////
      -entry.constant
    }
  }

  private def rankingFunction(rankingSolution : Conjunction,
                              rankingCoeffs : List[ConstantTerm],
                              stateConsts : List[ConstantTerm],
                              order : TermOrder) =
    LinearCombination(
      for ((coeffConst, stateConst) <- rankingCoeffs.elements zip stateConsts.elements)
        yield (extractValue(rankingSolution, coeffConst) getOrElse IdealInt.ZERO,
               stateConst),
      order)
}

////////////////////////////////////////////////////////////////////////////////

class RankingFinder(transition : Conjunction,
                    prePostConsts : List[(ConstantTerm, ConstantTerm)]) {

  import Conjunction._
  
  //////////////////////////////////////////////////////////////////////////////
  ApDebug.assertCtor(RankingFinder.AC, {
                     val preConsts = Set() ++ (for ((a, _) <- prePostConsts) yield a)
                     val postConsts = Set() ++ (for ((_, a) <- prePostConsts) yield a)
                     val allConsts = preConsts ++ postConsts

                     // the given variables for the pre-/post-state should all
                     // be distinct
                     preConsts.size == prePostConsts.size &&
                     postConsts.size == prePostConsts.size &&
                     allConsts.size == 2 * prePostConsts.size &&
                     // the transition relation has to be a quantifier-free
                     // Presburger formula
                     transition.predicates.isEmpty &&
                     transition.variables.isEmpty &&
                     (transition.constants subsetOf allConsts) &&
                     Conjunction.collectQuantifiers(transition).isEmpty})
  //////////////////////////////////////////////////////////////////////////////

  // We store the final ranking function, as well as a lower bound on the
  // value of the ranking function over the enabled initial states of the
  // transition relation. This bound is usually not a lower bound on the
  // post-state values of the transition function. In contrast, the
  // <code>postStateLowerBound</code> is.
  val (rankingFunction, preStateLowerBound, newConstants, postStateLowerBound) = {
    val order = transition.order
    
    print("Flattening ... ")

    // we first convert to DNF and eliminate negated equations
    val eqLessDisjuncts = 
      (for (oriDisjunct <- PresburgerTools.nonDNFEnumDisjuncts(transition);
            val reducedDisjunct =
              ReduceWithConjunction(Conjunction.TRUE, order)(oriDisjunct);
            eqLessDisjunct <- RankingFinder.splitEqs(reducedDisjunct))
       yield eqLessDisjunct).toList.removeDuplicates

    val varSplitter = new VariableSplitter(eqLessDisjuncts, prePostConsts, order)
    import varSplitter.{newOrder, newPrePostConsts}

    val disjuncts =
      (for (varSplitDisjunct <- varSplitter.newDisjuncts.elements;
            divLessDisjunct <- RankingFinder.splitDivisibilities(varSplitDisjunct)
            ) yield divLessDisjunct).toList.removeDuplicates
  
    println("" + disjuncts.size + " disjuncts")

    print("Generating constraints ... ")
    val (rankingCoeffs, lowerBound, constraint) =
      RankingFinder.setupRankingConstraints(disjuncts, newPrePostConsts, newOrder) 
    println("done")

    val solution = RankingFinder.minimalSolution(constraint)
    
    if (solution.isFalse) {
      (None, None, Map[ConstantTerm, RankingFinder.NewConstantDef](), None)
    } else {
      val preRankingFun =
        RankingFinder.rankingFunction(solution, rankingCoeffs,
                                      newPrePostConsts map (_._1), newOrder)
      val postRankingFun =
        RankingFinder.rankingFunction(solution, rankingCoeffs,
                                      newPrePostConsts map (_._2), newOrder)
      
      val absoluteLowerBound =
        PresburgerTools.lowerBound(postRankingFun, disj(disjuncts, newOrder)).get
      
      (Some(preRankingFun),
       Some(RankingFinder.extractValue(solution, lowerBound).get),
       Map() ++ varSplitter.newConstantsMap,
       Some(absoluteLowerBound))
    }
  }
  
}

////////////////////////////////////////////////////////////////////////////////

private object VariableSplitter {

  private def AC = Debug.AC_RANKING_FINDER

}

/**
 * A method to get rid of divisibility constraints. Whenever an expression
 * <code>EX modulus &times; _0 = t</code> occurs, we split the expression
 * <code>t</code> into two constants <code>c1, c2</code> that are defined by
 * <code>t = modulus &times; c1 + c2 & c2 >= 0 & c2 < modulus</code>.
 * Afterwards, it is much more efficient to eliminate the quantifier.
 */
private class VariableSplitter(disjuncts : Collection[Conjunction],
                               prePostConsts : List[(ConstantTerm, ConstantTerm)],
                               order : TermOrder) {
  
  private val preConsts = Set() ++ (for ((a, _) <- prePostConsts) yield a)
  private val postConsts = Set() ++ (for ((_, a) <- prePostConsts) yield a)

  import Conjunction._

  private val toPreSubst =
    ConstantSubst(Map() ++ (for ((pre, post) <- prePostConsts.elements)
                                 yield (post -> pre)), order)
  private val toPostSubst =
    ConstantSubst(Map() ++ (for ((pre, post) <- prePostConsts.elements)
                                 yield (pre -> post)), order)
  
  private val splitVariables =
    new scala.collection.mutable.HashMap[(IdealInt, LinearCombination),
                                         (ConstantTerm, ConstantTerm)]
  val newConstantsMap =
    new scala.collection.mutable.HashMap[ConstantTerm, RankingFinder.NewConstantDef]
  private var newConstsDefs = Conjunction.TRUE

  var newOrder = order
  
  private val splitDisjuncts =
    for (d <- disjuncts) yield splitVariables(d)

  val newDisjuncts =
    for (d <- splitDisjuncts) yield conj(Array(d, newConstsDefs), newOrder)

  val newPrePostConsts =
    prePostConsts ++ (for ((_, (preC, postC)) <- splitVariables) yield (preC, postC))

  //////////////////////////////////////////////////////////////////////////////
  
  private def splitVariables(c : Conjunction) : Conjunction = {
    ////////////////////////////////////////////////////////////////////////////
    ApDebug.assertPre(VariableSplitter.AC,
                      PresburgerTools.isQFPresburgerConjunction(c) &&
                      (c isSortedBy order))
    ////////////////////////////////////////////////////////////////////////////

    val simplifiedEquations = for (d <- c.negatedConjs) yield {
      //////////////////////////////////////////////////////////////////////////
      ApDebug.assertInt(VariableSplitter.AC, d.isNonDivisibility)
      //////////////////////////////////////////////////////////////////////////
      
      val ac = conj(c.arithConj, order)
      val matrixLC = d.arithConj.negativeEqs(0)
      val modulus = matrixLC.leadingCoeff
         
      // possibly split the groups of pre-state and of post-state variables
      
      val newExpr1 = splitSummand(matrixLC, preConsts, modulus, ac)
      val newExpr2 = splitSummand(newExpr1, postConsts, modulus, ac)

      d.updateNegativeEqs(NegEquationConj(newExpr2, newOrder))(newOrder)
    }
    
    c.updateNegatedConjs(NegatedConjunctions(simplifiedEquations, newOrder))(newOrder)
  }

  /**
   * Split those summands in <code>lc</code> that are built from constants in
   * <code>splitTerms</code>
   */
  private def splitSummand(lc : LinearCombination,
                           splitTerms : Set[ConstantTerm],
                           modulus : IdealInt,
                           constraints : Conjunction) : LinearCombination = {
    val lcToSplit =
      LinearCombination(
        FilterIt(lc.elements,
                 (x:(IdealInt,Term)) => !Seqs.disjoint((x _2).constants, splitTerms)),
        newOrder)
    ////////////////////////////////////////////////////////////////////////
    ApDebug.assertInt(VariableSplitter.AC,
                      (lcToSplit.constants subsetOf splitTerms) &&
                      lcToSplit.variables.isEmpty)
    ////////////////////////////////////////////////////////////////////////

    if (lcToSplit.isZero) {
      lc
    } else {
      val posLcToSplit = lcToSplit.makePositive
      
      PresburgerTools.bounds(posLcToSplit, constraints) match {
        case Some((lowerBound, upperBound))
        if (upperBound - lowerBound + 1 <= modulus) =>
          // if the known bounds on the value of the term are already tight
          // enough, we do not have to split
          lc
        case _ => {
          val newExpr = addConsts(posLcToSplit, modulus)
          LinearCombination.sum(Array((IdealInt.ONE, lc),
                                      (IdealInt.MINUS_ONE, lcToSplit),
                                      (IdealInt(lcToSplit.leadingCoeff.signum),
                                                                    newExpr)),
                                newOrder)
        }
      }
    }
  }
  
  //////////////////////////////////////////////////////////////////////////////
  
  /**
   * Bound the values of a constant to the range <code>[0, modulus)</code> by
   * introducing a new constant <code>c</code> that is defined by the
   * inequalities <code>0 <= t - modulus &times; c < modulus</code>. The result of
   * the method is the term <code>t - modulus &times; c</code>.
   */
  private def addConsts(oriExpr : LinearCombination,
                        modulus : IdealInt) : LinearCombination = {
    ////////////////////////////////////////////////////////////////////////////
    ApDebug.assertPre(VariableSplitter.AC,
                      (oriExpr isSortedBy order) &&
                      !oriExpr.constants.isEmpty &&
                      ((oriExpr.constants subsetOf preConsts) ||
                         (oriExpr.constants subsetOf postConsts)) &&
                      modulus.signum > 0)
    ////////////////////////////////////////////////////////////////////////////
    
    val preExpr = toPreSubst(oriExpr)
    
    val (preC, postC) = splitVariables.getOrElseUpdate((modulus, preExpr), {
      val postExpr = toPostSubst(oriExpr)

      val preC = new ConstantTerm ("t" + splitVariables.size + "_div_" + modulus)
      val postC = new ConstantTerm ("t" + splitVariables.size + "_div_" + modulus + "'")
      
      newOrder = newOrder.extend(preC, Set())
      newOrder = newOrder.extend(postC, Set())
      
      newConstantsMap += (preC -> RankingFinder.NewDivConstant(preExpr, modulus))
      newConstantsMap += (postC -> RankingFinder.NewDivConstant(postExpr, modulus))
      
      val defs = conj(
        for ((c1, expr) <- List((preC, preExpr), (postC, postExpr))) yield {
          val tMinusC =
            LinearCombination(Array((IdealInt.ONE, expr), (-modulus, c1)), newOrder)
          val upperBound =
            LinearCombination(Array((modulus - 1, OneTerm),
                                    (IdealInt.MINUS_ONE, tMinusC)), newOrder)
          conj(InEqConj(Array(tMinusC, upperBound), newOrder), newOrder)
        }, newOrder)
      
      newConstsDefs = conj(Array(newConstsDefs, defs), newOrder)
      
      (preC, postC)
    })
    
    LinearCombination(Array((IdealInt.ONE, oriExpr),
                            (-modulus,
                             if (oriExpr.constants subsetOf preConsts) preC else postC)),
                      newOrder)
  }
}